from typing import Optional, Tuple

import trio


async def connect_tcp(
    host: str,
    port: int,
    local_addr: Optional[Tuple[str, int]] = None,
) -> trio.SocketStream:
    return await trio.open_tcp_stream(
        host=host,
        port=port,
        local_address=local_addr,
    )



import socket


def is_connectable(host, port):
    sock = None
    try:
        sock = socket.create_connection((host, port), 1)
        result = True
    except socket.error:
        result = False
    finally:
        if sock:
            sock.close()
    return result



import enum
import ipaddress
import typing
from .._helpers import is_ip_address
from .._errors import ProxyError

RSV = NULL = 0x00
SOCKS_VER = 0x05


class AuthMethod(enum.IntEnum):
    ANONYMOUS = 0x00
    GSSAPI = 0x01
    USERNAME_PASSWORD = 0x02
    NO_ACCEPTABLE = 0xff


class AddressType(enum.IntEnum):
    IPV4 = 0x01
    DOMAIN = 0x03
    IPV6 = 0x04

    @classmethod
    def from_ip_ver(cls, ver: int):
        if ver == 4:
            return cls.IPV4
        if ver == 6:
            return cls.IPV6

        raise ValueError('Invalid IP version')


class Command(enum.IntEnum):
    CONNECT = 0x01
    BIND = 0x02
    UDP_ASSOCIATE = 0x03


class ReplyCode(enum.IntEnum):
    GRANTED = 0x00
    GENERAL_FAILURE = 0x01
    CONNECTION_NOT_ALLOWED = 0x02
    NETWORK_UNREACHABLE = 0x03
    HOST_UNREACHABLE = 0x04
    CONNECTION_REFUSED = 0x05
    TTL_EXPIRED = 0x06
    COMMAND_NOT_SUPPORTED = 0x07
    ADDRESS_TYPE_NOT_SUPPORTED = 0x08


ReplyMessages = {
    ReplyCode.GRANTED: 'Request granted',
    ReplyCode.GENERAL_FAILURE: 'General SOCKS server failure',
    ReplyCode.CONNECTION_NOT_ALLOWED: 'Connection not allowed by ruleset',
    ReplyCode.NETWORK_UNREACHABLE: 'Network unreachable',
    ReplyCode.HOST_UNREACHABLE: 'Host unreachable',
    ReplyCode.CONNECTION_REFUSED: 'Connection refused by destination host',
    ReplyCode.TTL_EXPIRED: 'TTL expired',
    ReplyCode.COMMAND_NOT_SUPPORTED: 'Command not supported or protocol error',
    ReplyCode.ADDRESS_TYPE_NOT_SUPPORTED: 'Address type not supported'
}


class AuthMethodsRequest:
    def __init__(self, username: str, password: str):
        auth_methods = bytearray([AuthMethod.ANONYMOUS])

        if username and password:
            auth_methods.append(AuthMethod.USERNAME_PASSWORD)

        self.auth_methods = auth_methods

    def __bytes__(self):
        return bytes([SOCKS_VER, len(self.auth_methods)]) + self.auth_methods


class AuthMethodsResponse:
    socks_ver: int
    auth_method: AuthMethod

    def __init__(self, data: bytes):
        assert len(data) == 2
        self.socks_ver = data[0]
        self.auth_method = data[1]  # noqa

    def validate(self, request: AuthMethodsRequest):
        if self.socks_ver != SOCKS_VER:
            raise ProxyError('Unexpected '  # pragma: no cover
                             'SOCKS version number: '
                             '{}'.format(self.socks_ver))

        if self.auth_method == AuthMethod.NO_ACCEPTABLE:
            raise ProxyError('No acceptable '  # pragma: no cover
                             'authentication methods were offered')

        if self.auth_method not in request.auth_methods:
            raise ProxyError('Unexpected SOCKS '  # pragma: no cover
                             'authentication method: '
                             '{}'.format(self.auth_method))


class AuthRequest(typing.SupportsBytes):
    VER = 0x01

    def __init__(self, username: str, password: str):
        self.username = username
        self.password = password

    def __bytes__(self):
        data = bytearray()
        data.append(self.VER)
        data.append(len(self.username))
        data += self.username.encode('ascii')
        data.append(len(self.password))
        data += self.password.encode('ascii')
        return bytes(data)


class AuthResponse:
    ver: int
    reply: ReplyCode

    def __init__(self, data: bytes):
        assert len(data) == 2
        self.ver = data[0]
        self.reply = data[1]  # noqa

    def validate(self):
        if self.ver != AuthRequest.VER:
            raise ProxyError('Invalid '  # pragma: no cover
                             'authentication response')

        if self.reply != ReplyCode.GRANTED:
            raise ProxyError('Username and password '  # pragma: no cover
                             'authentication failure')


class ConnectRequest:
    def __init__(self, host: str, port: int, rdns: bool):
        self.host = host
        self.port = port
        self.rdns = rdns
        self._resolved_host = None

    def __bytes__(self):
        data = bytearray([SOCKS_VER, Command.CONNECT, RSV])
        data += self._build_addr_request()
        return bytes(data)

    @property
    def need_resolve(self):
        return not is_ip_address(self.host) and not self.rdns

    def set_resolved_host(self, value):
        self._resolved_host = value

    def _build_addr_request(self) -> bytes:
        port = self.port.to_bytes(2, 'big')

        # destination address provided is an IPv4 or IPv6 address
        if is_ip_address(self.host):
            ip = ipaddress.ip_address(self.host)
            address_type = AddressType.from_ip_ver(ip.version)
            return bytes([address_type]) + ip.packed + port

        # not IP address, probably a DNS name
        if self.rdns:
            # resolve remotely
            address_type = AddressType.DOMAIN
            host = self.host.encode('idna')
            host_len = len(host)
            return bytes([address_type, host_len]) + host + port
        else:
            assert self._resolved_host is not None
            addr = self._resolved_host
            ip = ipaddress.ip_address(addr)
            address_type = AddressType.from_ip_ver(ip.version)
            return bytes([address_type]) + ip.packed + port


class ConnectResponse:
    socks_ver: int
    reply: ReplyCode
    rsv: int

    def __init__(self, data: bytes):
        assert len(data) == 3
        self.socks_ver = data[0]
        self.reply = data[1]  # noqa
        self.rsv = data[2]

    def validate(self):
        if self.socks_ver != SOCKS_VER:
            raise ProxyError('Unexpected SOCKS '  # pragma: no cover
                             'version number: {:#02X}'.format(self.socks_ver))

        if self.reply != ReplyCode.GRANTED:  # pragma: no cover
            msg = ReplyMessages.get(self.reply, 'Unknown error')
            raise ProxyError(msg, self.reply)

        if self.rsv != RSV:
            raise ProxyError('The reserved byte '  # pragma: no cover
                             'must be {:#02X}'.format(RSV))



import sys

from .._basic_auth import BasicAuth
from .._errors import ProxyError
from .._version import __title__, __version__

DEFAULT_USER_AGENT = 'Python/{0[0]}.{0[1]} {1}/{2}'.format(
    sys.version_info, __title__, __version__)

CRLF = '\r\n'


class ConnectRequest:
    def __init__(self, host: str, port: int, login: str, password: str):
        self.host = host
        self.port = port
        self.login = login
        self.password = password

    def __bytes__(self):
        req = [
            'CONNECT {}:{} HTTP/1.1'.format(self.host, self.port),
            'Host: {}:{}'.format(self.host, self.port),
            'User-Agent: {}'.format(DEFAULT_USER_AGENT),
        ]

        if self.login and self.password:
            auth = BasicAuth(self.login, self.password)
            req.append('Proxy-Authorization: {}'.format(auth.encode()))

        req.append(CRLF)

        data = CRLF.join(req).encode('ascii')
        return data


class ConnectResponse:
    def __init__(self, data: bytes):
        self.data = data

    def validate(self):
        data = self.data

        if not data:
            raise ProxyError('Invalid proxy response')  # pragma: no cover'

        line = data.split(CRLF.encode('ascii'), 1)[0]
        line = line.decode('utf-8', 'surrogateescape')

        try:
            version, code, *reason = line.split()
        except ValueError:  # pragma: no cover
            raise ProxyError('Invalid status line: {}'.format(line))

        try:
            status_code = int(code)
        except ValueError:  # pragma: no cover
            raise ProxyError('Invalid status code: {}'.format(code))

        if status_code != 200:
            msg = 'Proxy error: {} {}'.format(status_code, ' '.join(reason))
            raise ProxyError(msg, status_code)



from contextlib import contextmanager
from unittest import mock

import pytest

from python_socks.async_.asyncio._resolver import Resolver as AsyncioResolver
from python_socks.sync._resolver import SyncResolver
from tests.config import (
    PROXY_HOST_IPV4,
    PROXY_HOST_IPV6,
    SOCKS5_PROXY_PORT,
    LOGIN,
    PASSWORD,
    SKIP_IPV6_TESTS,
    HTTP_PROXY_PORT,
    SOCKS4_PORT_NO_AUTH,
    SOCKS4_PROXY_PORT,
    SOCKS5_PROXY_PORT_NO_AUTH,
    TEST_PORT_IPV4,
    TEST_PORT_IPV6,
    TEST_HOST_IPV4,
    TEST_HOST_IPV6,
    TEST_PORT_IPV4_HTTPS,
    TEST_HOST_CERT_FILE,
    TEST_HOST_KEY_FILE,
    HTTPS_PROXY_PORT, PROXY_HOST_CERT_FILE, PROXY_HOST_KEY_FILE,
)
from tests.http_server import HttpServer, HttpServerConfig
from tests.mocks import sync_resolve_factory, async_resolve_factory
from tests.proxy_server import ProxyConfig, ProxyServer


@contextmanager
def nullcontext():
    yield None


# @pytest.fixture(scope='session', autouse=True)
# def patch_socket_getaddrinfo():
#     with mock.patch('socket.getaddrinfo', side_effect=getaddrinfo):
#         yield None


@pytest.fixture(scope='session', autouse=True)
def patch_resolvers():
    p1 = mock.patch.object(
        SyncResolver, attribute='resolve', new=sync_resolve_factory(SyncResolver)
    )

    p2 = mock.patch.object(
        AsyncioResolver, attribute='resolve', new=async_resolve_factory(AsyncioResolver)
    )

    try:
        # noinspection PyProtectedMember
        from python_socks.async_.trio._resolver import Resolver as TrioResolver
    except ImportError:
        p3 = nullcontext()
    else:
        p3 = mock.patch.object(
            TrioResolver, attribute='resolve', new=async_resolve_factory(TrioResolver)
        )

    try:
        # noinspection PyProtectedMember
        from python_socks.async_.curio._resolver import Resolver as CurioResolver
    except ImportError:
        p4 = nullcontext()
    else:
        p4 = mock.patch.object(
            CurioResolver, attribute='resolve', new=async_resolve_factory(CurioResolver)
        )

    try:
        from python_socks.async_.anyio._resolver import Resolver as AnyioResolver
    except ImportError:
        p5 = nullcontext()
    else:
        p5 = mock.patch.object(
            AnyioResolver, attribute='resolve', new=async_resolve_factory(AnyioResolver)
        )

    with p1, p2, p3, p4, p5:
        yield None


@pytest.fixture(scope='session', autouse=True)
def proxy_server():
    config = [
        ProxyConfig(
            proxy_type='http',
            host=PROXY_HOST_IPV4,
            port=HTTP_PROXY_PORT,
            username=LOGIN,
            password=PASSWORD,
        ),
        ProxyConfig(
            proxy_type='socks4',
            host=PROXY_HOST_IPV4,
            port=SOCKS4_PROXY_PORT,
            username=LOGIN,
            password=None,
        ),
        ProxyConfig(
            proxy_type='socks4',
            host=PROXY_HOST_IPV4,
            port=SOCKS4_PORT_NO_AUTH,
            username=None,
            password=None,
        ),
        ProxyConfig(
            proxy_type='socks5',
            host=PROXY_HOST_IPV4,
            port=SOCKS5_PROXY_PORT,
            username=LOGIN,
            password=PASSWORD,
        ),
        ProxyConfig(
            proxy_type='socks5',
            host=PROXY_HOST_IPV4,
            port=SOCKS5_PROXY_PORT_NO_AUTH,
            username=None,
            password=None,
        ),
        ProxyConfig(
            proxy_type='http',
            # host=PROXY_HOST_NAME_IPV4,
            host=PROXY_HOST_IPV4,
            port=HTTPS_PROXY_PORT,
            username=LOGIN,
            password=PASSWORD,
            certfile=PROXY_HOST_CERT_FILE,
            keyfile=PROXY_HOST_KEY_FILE,
        ),
    ]

    if not SKIP_IPV6_TESTS:
        config.append(
            ProxyConfig(
                proxy_type='socks5',
                host=PROXY_HOST_IPV6,
                port=SOCKS5_PROXY_PORT,
                username=LOGIN,
                password=PASSWORD,
            ),
        )

    server = ProxyServer(config=config)
    server.start()
    for cfg in config:
        server.wait_until_connectable(host=cfg.host, port=cfg.port)

    yield None

    server.terminate()


@pytest.fixture(scope='session', autouse=True)
def web_server():
    config = [
        HttpServerConfig(
            host=TEST_HOST_IPV4,
            port=TEST_PORT_IPV4,
        ),
        HttpServerConfig(
            host=TEST_HOST_IPV4,
            port=TEST_PORT_IPV4_HTTPS,
            certfile=TEST_HOST_CERT_FILE,
            keyfile=TEST_HOST_KEY_FILE,
        ),
    ]

    if not SKIP_IPV6_TESTS:
        config.append(HttpServerConfig(host=TEST_HOST_IPV6, port=TEST_PORT_IPV6))

    server = HttpServer(config=config)
    server.start()
    for cfg in config:
        server.wait_until_connectable(host=cfg.host, port=cfg.port)

    yield None

    server.terminate()



import socket
from typing import Optional, Tuple


def connect_tcp(
    host: str,
    port: int,
    timeout: float = None,
    local_addr: Optional[Tuple[str, int]] = None,
) -> socket.socket:
    address = (host, port)
    return socket.create_connection(
        address,
        timeout,
        source_address=local_addr,
    )



import asyncio
import socket
import sys
from typing import Optional

import async_timeout

from ..._errors import ProxyConnectionError, ProxyTimeoutError
from ..._proto.http_async import HttpProto
from ..._proto.socks4_async import Socks4Proto
from ..._proto.socks5_async import Socks5Proto
from ._stream import AsyncioSocketStream
from ._resolver import Resolver

from ._connect import connect_tcp
from ... import _abc as abc

DEFAULT_TIMEOUT = 60


class AsyncioProxy(abc.AsyncProxy):
    _stream: Optional[abc.AsyncSocketStream]

    def __init__(
        self,
        proxy_host: str,
        proxy_port: int,
        loop: asyncio.AbstractEventLoop = None,
    ):

        if loop is None:
            loop = asyncio.get_event_loop()

        self._loop = loop

        self._proxy_host = proxy_host
        self._proxy_port = proxy_port

        self._dest_host = None
        self._dest_port = None
        self._timeout = None

        self._stream = None
        self._resolver = Resolver(loop=loop)

    async def connect(
        self,
        dest_host: str,
        dest_port: int,
        timeout: float = None,
        _socket=None,
    ) -> socket.socket:

        if timeout is None:
            timeout = DEFAULT_TIMEOUT

        self._dest_host = dest_host
        self._dest_port = dest_port
        self._timeout = timeout

        try:
            return await self._connect(_socket=_socket)
        except asyncio.TimeoutError as e:
            raise ProxyTimeoutError('Proxy connection timed out: {}'.format(self._timeout)) from e

    async def _connect(self, _socket=None) -> socket.socket:
        async with async_timeout.timeout(self._timeout):
            try:
                if _socket is None:
                    _socket = await connect_tcp(
                        host=self._proxy_host,
                        port=self._proxy_port,
                        loop=self._loop,
                    )

                self._stream = AsyncioSocketStream(sock=_socket, loop=self._loop)
                await self._negotiate()
                return _socket
            except OSError as e:
                await self._close()
                msg = 'Could not connect to proxy {}:{} [{}]'.format(
                    self._proxy_host,
                    self._proxy_port,
                    e.strerror,
                )
                raise ProxyConnectionError(e.errno, msg) from e
            except asyncio.CancelledError:  # pragma: no cover
                # https://bugs.python.org/issue30064
                # https://bugs.python.org/issue34795
                if self._can_be_closed_safely():
                    await self._close()
                raise
            except Exception:  # pragma: no cover
                await self._close()
                raise

    def _can_be_closed_safely(self):  # pragma: no cover
        def is_proactor_event_loop():
            try:
                from asyncio import ProactorEventLoop  # noqa
            except ImportError:
                return False
            return isinstance(self._loop, ProactorEventLoop)

        def is_uvloop_event_loop():
            try:
                from uvloop import Loop  # noqa
            except ImportError:
                return False
            return isinstance(self._loop, Loop)

        return sys.version_info[:2] >= (3, 8) or is_proactor_event_loop() or is_uvloop_event_loop()

    async def _negotiate(self):
        raise NotImplementedError()  # pragma: no cover

    async def _close(self):
        if self._stream is not None:
            await self._stream.close()

    @property
    def proxy_host(self):
        return self._proxy_host

    @property
    def proxy_port(self):
        return self._proxy_port


class Socks5Proxy(AsyncioProxy):
    def __init__(
        self,
        proxy_host,
        proxy_port,
        username=None,
        password=None,
        rdns=None,
        loop: asyncio.AbstractEventLoop = None,
    ):
        super().__init__(proxy_host=proxy_host, proxy_port=proxy_port, loop=loop)
        self._username = username
        self._password = password
        self._rdns = rdns

    async def _negotiate(self):
        proto = Socks5Proto(
            stream=self._stream,
            resolver=self._resolver,
            dest_host=self._dest_host,
            dest_port=self._dest_port,
            username=self._username,
            password=self._password,
            rdns=self._rdns,
        )
        await proto.negotiate()


class Socks4Proxy(AsyncioProxy):
    def __init__(
        self,
        proxy_host,
        proxy_port,
        user_id=None,
        rdns=None,
        loop: asyncio.AbstractEventLoop = None,
    ):
        super().__init__(proxy_host=proxy_host, proxy_port=proxy_port, loop=loop)
        self._user_id = user_id
        self._rdns = rdns

    async def _negotiate(self):
        proto = Socks4Proto(
            stream=self._stream,
            resolver=self._resolver,
            dest_host=self._dest_host,
            dest_port=self._dest_port,
            user_id=self._user_id,
            rdns=self._rdns,
        )
        await proto.negotiate()


class HttpProxy(AsyncioProxy):
    def __init__(
        self,
        proxy_host,
        proxy_port,
        username=None,
        password=None,
        loop: asyncio.AbstractEventLoop = None,
    ):
        super().__init__(proxy_host=proxy_host, proxy_port=proxy_port, loop=loop)
        self._username = username
        self._password = password

    async def _negotiate(self):
        proto = HttpProto(
            stream=self._stream,
            dest_host=self._dest_host,
            dest_port=self._dest_port,
            username=self._username,
            password=self._password,
        )
        await proto.negotiate()



import socket

from .socks4 import ConnectRequest, ConnectResponse
from .. import _abc as abc


class Socks4Proto:
    def __init__(
        self,
        stream: abc.AsyncSocketStream,
        resolver: abc.AsyncResolver,
        dest_host,
        dest_port,
        user_id=None,
        rdns=None,
    ):

        if rdns is None:
            rdns = False

        self._dest_host = dest_host
        self._dest_port = dest_port

        self._user_id = user_id
        self._rdns = rdns

        self._stream = stream
        self._resolver = resolver

    async def negotiate(self):
        await self._socks_connect()

    async def _socks_connect(self):
        req = ConnectRequest(
            host=self._dest_host,
            port=self._dest_port,
            user_id=self._user_id,
            rdns=self._rdns,
        )

        if req.need_resolve:
            _, addr = await self._resolver.resolve(req.host, family=socket.AF_INET)
            req.set_resolved_host(addr)

        await self._stream.write_all(bytes(req))

        data = await self._stream.read_exact(8)

        res = ConnectResponse(data[:2])
        res.validate()



import asyncio
from typing import Optional, Tuple


async def connect_tcp(
    host: str,
    port: int,
    local_addr: Optional[Tuple[str, int]] = None,
) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:

    kwargs = {}
    if local_addr is not None:
        kwargs['local_addr'] = local_addr   # pragma: no cover

    reader, writer = await asyncio.open_connection(
        host=host,
        port=port,
        **kwargs,
    )

    return reader, writer



from ...._proxy_factory import ProxyFactory
from ...._types import ProxyType
from ._proxy import Socks5Proxy, Socks4Proxy, HttpProxy, TrioProxy


class Proxy(ProxyFactory[TrioProxy]):
    types = {
        ProxyType.SOCKS4: Socks4Proxy,
        ProxyType.SOCKS5: Socks5Proxy,
        ProxyType.HTTP: HttpProxy,
    }



import logging
import ssl
import sys
import time
import typing
from multiprocessing import Process
from unittest import mock

from tests.mocks import sync_resolve_factory
from tests.tiny_proxy import HttpProxyServer, Socks4ProxyServer, Socks5ProxyServer
from tests.tiny_proxy.handlers import (
    Socks4ProxyHandler,
    Socks5ProxyHandler,
    HttpProxyHandler,
)
from tests.tiny_proxy.handlers.base_proxy import BaseProxyHandler
from tests.tiny_proxy.handlers.resolver import Resolver
from tests.utils import is_connectable


class ProxyConfig(typing.NamedTuple):
    proxy_type: str
    host: str
    port: int
    username: typing.Optional[str] = None
    password: typing.Optional[str] = None
    certfile: typing.Optional[str] = None
    keyfile: typing.Optional[str] = None

    def to_dict(self):
        d = {}
        for key, val in self._asdict().items():
            if val is not None:
                d[key] = val
        return d


cls_map = {
    'http': HttpProxyServer,
    'socks4': Socks4ProxyServer,
    'socks5': Socks5ProxyServer,
}


def configure_logging():
    root_logger = logging.getLogger()
    root_logger.setLevel('DEBUG')

    fmt = '%(asctime)s [%(name)s] %(levelname)s : %(message)s'
    formatter = logging.Formatter(fmt)
    formatter.converter = time.gmtime

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(formatter)
    root_logger.addHandler(stdout_handler)


def connect_to_remote_factory(cls: typing.Type[BaseProxyHandler]):
    """
    simulate target host connection timeout
    """
    origin_connect_to_remote = cls.connect_to_remote

    def new_connect_to_remote(self):
        time.sleep(0.01)
        return origin_connect_to_remote(self)

    return new_connect_to_remote


@mock.patch.object(
    HttpProxyHandler,
    attribute='connect_to_remote',
    new=connect_to_remote_factory(HttpProxyHandler),
)
@mock.patch.object(
    Socks4ProxyHandler,
    attribute='connect_to_remote',
    new=connect_to_remote_factory(Socks4ProxyHandler),
)
@mock.patch.object(
    Socks5ProxyHandler,
    attribute='connect_to_remote',
    new=connect_to_remote_factory(Socks5ProxyHandler),
)
@mock.patch.object(Resolver, attribute='resolve', new=sync_resolve_factory(Resolver))
def start(proxy_type, host, port, certfile=None, keyfile=None, **kwargs):
    # configure_logging()

    cls = cls_map.get(proxy_type)
    if not cls:
        raise RuntimeError('Unsupported type: {}'.format(proxy_type))

    if certfile is not None and keyfile is not None:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        ssl_context.load_cert_chain(certfile, keyfile)
    else:
        ssl_context = None

    with cls(host, port, ssl_context=ssl_context, **kwargs) as server:
        server.serve_forever()


class ProxyServer:
    workers: typing.List[Process]

    def __init__(self, config: typing.Iterable[ProxyConfig]):
        self.config = config
        self.workers = []

    def start(self):
        for cfg in self.config:
            print(
                'Starting {} proxy on {}:{}; certfile={}, keyfile={}...'.format(
                    cfg.proxy_type,
                    cfg.host,
                    cfg.port,
                    cfg.certfile,
                    cfg.keyfile,
                )
            )

            p = Process(target=start, kwargs=cfg.to_dict())
            self.workers.append(p)

        for p in self.workers:
            p.start()

    def terminate(self):
        for p in self.workers:
            p.terminate()

    def wait_until_connectable(self, host, port, timeout=10):
        count = 0
        while not is_connectable(host=host, port=port):
            if count >= timeout:
                self.terminate()
                raise Exception(
                    'The proxy server has not available by (%s, %s) in %d seconds'
                    % (host, port, timeout)
                )
            count += 1
            time.sleep(1)
        return True



from typing import Iterable
from ._proxy import SyncProxy


class ProxyChain:
    def __init__(self, proxies: Iterable[SyncProxy]):
        self._proxies = proxies

    def connect(
        self,
        dest_host,
        dest_port,
        dest_ssl=None,
        timeout=None,
    ):
        stream = None
        proxies = list(self._proxies)

        length = len(proxies) - 1
        for i in range(length):
            stream = proxies[i].connect(
                dest_host=proxies[i + 1].proxy_host,
                dest_port=proxies[i + 1].proxy_port,
                timeout=timeout,
                _stream=stream,
            )

        stream = proxies[length].connect(
            dest_host=dest_host,
            dest_port=dest_port,
            dest_ssl=dest_ssl,
            timeout=timeout,
            _stream=stream,
        )

        return stream



from typing import Type, Dict, Any, TypeVar, Generic

from ._helpers import parse_proxy_url
from ._types import ProxyType

T = TypeVar('T')


class ProxyFactory(Generic[T]):
    types: Dict[ProxyType, Type[Any]]

    @classmethod
    def create(
        cls,
        proxy_type: ProxyType,
        host: str,
        port: int,
        username: str = None,
        password: str = None,
        rdns: bool = None,
        **kwargs,
    ) -> T:

        proxy_cls = cls.types.get(proxy_type)

        if proxy_type == ProxyType.SOCKS4:
            return proxy_cls(
                proxy_host=host, proxy_port=port, user_id=username, rdns=rdns, **kwargs
            )

        if proxy_type == ProxyType.SOCKS5:
            return proxy_cls(
                proxy_host=host,
                proxy_port=port,
                username=username,
                password=password,
                rdns=rdns,
                **kwargs,
            )

        if proxy_type == ProxyType.HTTP:
            return proxy_cls(
                proxy_host=host,
                proxy_port=port,
                username=username,
                password=password,
                **kwargs,
            )

        raise ValueError('Invalid proxy type: {}'.format(proxy_type))

    @classmethod
    def from_url(cls, url: str, **kwargs) -> T:
        proxy_type, host, port, username, password = parse_proxy_url(url)
        return cls.create(
            proxy_type=proxy_type,
            host=host,
            port=port,
            username=username,
            password=password,
            **kwargs,
        )



import logging
import select
import socket
from socketserver import StreamRequestHandler

DEFAULT_RECEIVE_SIZE = 65536


class ProxyError(Exception):
    pass


class BaseProxyHandler(StreamRequestHandler):
    request: socket.socket
    logger: logging.Logger

    def __init__(self, request, client_address, server, ):
        self.logger = logging.getLogger(__name__)
        super().__init__(request, client_address, server)

    def handle(self):
        client = self.request
        try:
            remote = self.connect_to_remote()
        except Exception as e:
            self.logger.error(e)
            self.logger.debug(e, exc_info=True)
            client.close()
        else:
            try:
                self.exchange_loop(client=client, remote=remote)
            except Exception as e:
                self.logger.exception(e)
            finally:
                remote.close()
                client.close()

    def connect_to_remote(self) -> socket.socket:
        raise NotImplementedError

    # noinspection PyMethodMayBeStatic
    def exchange_loop(self, client: socket.socket, remote: socket.socket):
        while True:

            # wait until client or remote is available for read
            r, w, e = select.select([client, remote], [], [])

            if client in r:
                data = client.recv(DEFAULT_RECEIVE_SIZE)
                if remote.send(data) <= 0:
                    break

            if remote in r:
                data = remote.recv(DEFAULT_RECEIVE_SIZE)
                if client.send(data) <= 0:
                    break



import functools
import ssl
import socket
from socketserver import ThreadingMixIn, TCPServer

from .handlers import HttpProxyHandler, Socks5ProxyHandler, Socks4ProxyHandler
from .helpers import is_ipv6_address


class ProxyServer(ThreadingMixIn, TCPServer):
    allow_reuse_address = True

    def __init__(self, addr, handler_cls, ssl_context: ssl.SSLContext = None):
        if is_ipv6_address(addr[0]):
            self.address_family = socket.AF_INET6

        super().__init__(server_address=addr, RequestHandlerClass=handler_cls)

        self.ssl_context = ssl_context

    def get_request(self):
        sock, addr = self.socket.accept()

        if self.ssl_context is not None:
            sock = self.ssl_context.wrap_socket(sock, server_side=True)

        return sock, addr


class HttpProxyServer(ProxyServer):
    def __init__(self, host: str, port: int,
                 username=None, password=None,
                 ssl_context: ssl.SSLContext = None):
        handler_cls = functools.partial(
            HttpProxyHandler,
            username=username,
            password=password
        )
        super().__init__(
            addr=(host, port),
            handler_cls=handler_cls,
            ssl_context=ssl_context
        )


class Socks5ProxyServer(ProxyServer):
    def __init__(self, host: str, port: int,
                 username=None, password=None,
                 ssl_context: ssl.SSLContext = None):
        handler_cls = functools.partial(
            Socks5ProxyHandler,
            username=username,
            password=password
        )
        super().__init__(
            addr=(host, port),
            handler_cls=handler_cls,
            ssl_context=ssl_context
        )


class Socks4ProxyServer(ProxyServer):
    def __init__(self, host: str, port: int,
                 username=None, ssl_context: ssl.SSLContext = None):
        handler_cls = functools.partial(
            Socks4ProxyHandler,
            username=username
        )
        super().__init__(
            addr=(host, port),
            handler_cls=handler_cls,
            ssl_context=ssl_context
        )



from .http_proxy import HttpProxyHandler
from .socks5_proxy import Socks5ProxyHandler
from .socks4_proxy import Socks4ProxyHandler

__all__ = (
    'HttpProxyHandler',
    'Socks5ProxyHandler',
    'Socks4ProxyHandler',
)



from ._chain import ProxyChain

from .._types import ProxyType
from .._proxy_factory import ProxyFactory
from ._proxy import Socks5Proxy, Socks4Proxy, HttpProxy, SyncProxy


class Proxy(ProxyFactory[SyncProxy]):
    types = {
        ProxyType.SOCKS4: Socks4Proxy,
        ProxyType.SOCKS5: Socks5Proxy,
        ProxyType.HTTP: HttpProxy,
    }


__all__ = ('Proxy', 'ProxyChain')



import enum
import ipaddress
import logging
import socket

from .base_proxy import BaseProxyHandler, ProxyError
from .resolver import Resolver
from ..helpers import is_ipv4_address, is_ipv6_address

RSV = NULL = 0x00
SOCKS_VER5 = 0x05

SOCKS5_GRANTED = 0x00


class AuthMethod(enum.IntEnum):
    ANONYMOUS = 0x00
    GSSAPI = 0x01
    USERNAME_PASSWORD = 0x02
    NO_ACCEPTABLE = 0xff


class AddressType(enum.IntEnum):
    IPV4 = 0x01
    DOMAIN = 0x03
    IPV6 = 0x04

    @classmethod
    def from_ip_ver(cls, ver: int):
        if ver == 4:
            return cls.IPV4
        if ver == 6:
            return cls.IPV6

        raise ValueError('Invalid IP version')


class Command(enum.IntEnum):
    CONNECT = 0x01
    BIND = 0x02
    UDP_ASSOCIATE = 0x03


class ReplyCode(enum.IntEnum):
    SUCCEEDED = 0x00
    GENERAL_FAILURE = 0x01
    CONNECTION_NOT_ALLOWED = 0x02
    NETWORK_UNREACHABLE = 0x03
    HOST_UNREACHABLE = 0x04
    CONNECTION_REFUSED = 0x05
    TTL_EXPIRED = 0x06
    COMMAND_NOT_SUPPORTED = 0x07
    ADDRESS_TYPE_NOT_SUPPORTED = 0x08


ReplyMessages = {
    ReplyCode.SUCCEEDED: 'Request granted',
    ReplyCode.GENERAL_FAILURE: 'General SOCKS server failure',
    ReplyCode.CONNECTION_NOT_ALLOWED: 'Connection not allowed by ruleset',
    ReplyCode.NETWORK_UNREACHABLE: 'Network unreachable',
    ReplyCode.HOST_UNREACHABLE: 'Host unreachable',
    ReplyCode.CONNECTION_REFUSED: 'Connection refused by destination host',
    ReplyCode.TTL_EXPIRED: 'TTL expired',
    ReplyCode.COMMAND_NOT_SUPPORTED: 'Command not supported or protocol error',
    ReplyCode.ADDRESS_TYPE_NOT_SUPPORTED: 'Address type not supported'
}


class Socks5ProxyHandler(BaseProxyHandler):
    def __init__(self, request, client_address, server, *,
                 username=None, password=None):
        self.username = username
        self.password = password
        self.logger = logging.getLogger(__name__)
        self.resolver = Resolver()

        super().__init__(request, client_address, server)

    def connect_to_remote(self) -> socket.socket:
        client = self.request

        auth_required = self.username and self.password

        # ------------------------AUTH METHODS---------------------
        version, num_methods = client.recv(2)

        if version != SOCKS_VER5:
            client.sendall(bytes([NULL, NULL]))
            raise ProxyError('Unsupported socks version')

        methods = []
        for i in range(num_methods):
            methods.append(ord(client.recv(1)))

        if auth_required:
            auth_method = (AuthMethod.USERNAME_PASSWORD
                           if AuthMethod.USERNAME_PASSWORD in methods
                           else AuthMethod.NO_ACCEPTABLE)
        else:
            auth_method = (AuthMethod.ANONYMOUS
                           if AuthMethod.ANONYMOUS in methods
                           else AuthMethod.NO_ACCEPTABLE)

        client.sendall(bytes([SOCKS_VER5, auth_method]))
        if auth_method == AuthMethod.NO_ACCEPTABLE:
            raise ProxyError('Not acceptable auth method')

        # -----------------------AUTH REQUEST-----------------------
        if auth_method == AuthMethod.USERNAME_PASSWORD:
            version = ord(client.recv(1))
            if version != 1:
                client.sendall(bytes([version, 0xFF]))
                raise ProxyError('Invalid auth request')

            username_len = ord(client.recv(1))
            username = client.recv(username_len).decode('utf-8')

            password_len = ord(client.recv(1))
            password = client.recv(password_len).decode('utf-8')

            if username == self.username and password == self.password:
                client.sendall(bytes([version, SOCKS5_GRANTED]))
            else:
                client.sendall(bytes([version, 0xFF]))
                raise ProxyError('Authentication failed')

        # --------------------------CONNECT-----------------------------
        version, cmd, _, address_type = client.recv(4)

        if version != SOCKS_VER5:
            client.sendall(bytes([
                SOCKS_VER5,
                ReplyCode.GENERAL_FAILURE,
                RSV
            ]))
            raise ProxyError(ReplyMessages[ReplyCode.GENERAL_FAILURE])

        if cmd != Command.CONNECT:
            client.sendall(bytes([
                SOCKS_VER5,
                ReplyCode.COMMAND_NOT_SUPPORTED,
                RSV
            ]))
            raise ProxyError(ReplyMessages[ReplyCode.COMMAND_NOT_SUPPORTED])

        if address_type == AddressType.IPV4:
            remote_host = str(ipaddress.IPv4Address(client.recv(4)))
        elif address_type == AddressType.IPV6:
            remote_host = str(ipaddress.IPv6Address(client.recv(16)))
        elif address_type == AddressType.DOMAIN:
            domain_length = ord(client.recv(1))
            remote_host = client.recv(domain_length).decode('ascii')
        else:
            client.sendall(bytes([
                SOCKS_VER5,
                ReplyCode.ADDRESS_TYPE_NOT_SUPPORTED,
                NULL, NULL, NULL, NULL
            ]))
            raise ProxyError(
                ReplyMessages[ReplyCode.ADDRESS_TYPE_NOT_SUPPORTED])

        remote_port = int.from_bytes(client.recv(2), 'big')

        self.logger.info('CONNECT {}:{}'.format(remote_host, remote_port))

        try:
            # remote = socket.create_connection((remote_host, remote_port))
            family, host = self.resolve_target_host(remote_host)

            remote = socket.socket(
                family=family,
                type=socket.SOCK_STREAM
            )

            remote.connect((host, remote_port))
        except Exception as e:
            self.logger.error(e)
            reply = bytes([
                SOCKS_VER5,
                ReplyCode.CONNECTION_REFUSED,
                NULL, NULL, NULL, NULL
            ])
            client.sendall(reply)
            raise ProxyError(
                ReplyMessages[ReplyCode.CONNECTION_REFUSED]) from e
        else:
            bind_address = remote.getsockname()
            bind_ip = ipaddress.ip_address(bind_address[0])
            bind_port = bind_address[1]

            reply = bytearray([
                SOCKS_VER5,
                ReplyCode.SUCCEEDED,
                RSV,
                AddressType.from_ip_ver(bind_ip.version)
            ])
            reply += bind_ip.packed
            reply += bind_port.to_bytes(2, 'big')

            client.sendall(reply)

            return remote

    def resolve_target_host(self, host):
        if is_ipv4_address(host):
            return socket.AF_INET, host
        if is_ipv6_address(host):
            return socket.AF_INET6, host
        return self.resolver.resolve(host=host)



import logging
import socket
from http.server import BaseHTTPRequestHandler
from io import BytesIO

from .auth import BasicAuth
from .base_proxy import BaseProxyHandler, ProxyError, DEFAULT_RECEIVE_SIZE
from .resolver import Resolver
from ..helpers import is_ipv4_address, is_ipv6_address


class HTTPRequest(BaseHTTPRequestHandler):
    """
    https://stackoverflow.com/questions/4685217/parse-raw-http-headers
    """

    # noinspection PyMissingConstructor
    def __init__(self, data: bytes):
        self.rfile = BytesIO(data)
        self.raw_requestline = self.rfile.readline()
        self.error_code = self.error_message = None
        self.parse_request()

    def send_error(self, code, message=None, explain=None):
        self.error_code = code


class HTTPResponse:
    def __init__(self, code: int, message: str):
        self.code = code
        self.message = message


class HttpProxyHandler(BaseProxyHandler):
    def __init__(self, request, client_address, server, *,
                 username=None, password=None):
        self.username = username
        self.password = password
        self.logger = logging.getLogger(__name__)
        self.resolver = Resolver()

        super().__init__(request, client_address, server)

    def connect_to_remote(self) -> socket.socket:
        client = self.request

        req = HTTPRequest(client.recv(DEFAULT_RECEIVE_SIZE))

        if req.error_code:
            self.respond(int(req.error_code), req.error_message)

        if req.command.lower() != 'connect':
            self.respond(400, 'Bad Request')

        if self.username and self.password:
            auth_header = req.headers['proxy-authorization']

            if not auth_header:
                self.respond(401, 'Unauthorized')

            try:
                auth = BasicAuth.decode(auth_header)
            except ValueError:
                self.respond(401, 'Unauthorized')
            else:
                if (auth.login != self.username
                        or auth.password != self.password):
                    self.respond(401, 'Unauthorized')

        try:
            host, port = req.path.split(':')
            port = int(port)
        except ValueError:
            self.respond(400, 'Bad Request')
            raise

        self.logger.info('CONNECT {}:{}'.format(host, port))
        try:
            # remote = socket.create_connection((host, port))
            family, host = self.resolve_target_host(host)

            remote = socket.socket(
                family=family,
                type=socket.SOCK_STREAM
            )

            remote.connect((host, port))
        except Exception as e:
            self.logger.error(e)
            self.respond(502, 'Bad Gateway')
        else:
            self.respond(200, 'Connection established')
            return remote

    def respond(self, code: int, message: str):
        res = 'HTTP/1.1 {} {}\r\n\r\n'.format(code, message)
        self.request.sendall(res.encode('ascii'))
        if code != 200:
            raise ProxyError('{} {}'.format(code, message))

    def resolve_target_host(self, host):
        if is_ipv4_address(host):
            return socket.AF_INET, host
        if is_ipv6_address(host):
            return socket.AF_INET6, host
        return self.resolver.resolve(host=host)



import socket

from .socks5 import (
    AuthMethod,
    AuthMethodsRequest,
    AuthMethodsResponse,
    AuthRequest,
    AuthResponse,
    ConnectRequest,
    ConnectResponse,
)
from .. import _abc as abc


class Socks5Proto:
    def __init__(
        self,
        stream: abc.AsyncSocketStream,
        resolver: abc.AsyncResolver,
        dest_host,
        dest_port,
        username=None,
        password=None,
        rdns=None,
    ):

        if rdns is None:
            rdns = True

        self._dest_port = dest_port
        self._dest_host = dest_host

        self._username = username
        self._password = password
        self._rdns = rdns

        self._stream = stream
        self._resolver = resolver

    async def negotiate(self):
        await self._socks_auth()
        await self._socks_connect()

    async def _socks_auth(self):
        auth_method = await self._choose_auth_method()

        # authenticate
        if auth_method == AuthMethod.USERNAME_PASSWORD:
            req = AuthRequest(username=self._username, password=self._password)

            await self._stream.write_all(bytes(req))

            res = AuthResponse(await self._stream.read_exact(2))
            res.validate()

    async def _choose_auth_method(self) -> AuthMethod:
        req = AuthMethodsRequest(
            username=self._username,
            password=self._password,
        )

        await self._stream.write_all(bytes(req))

        res = AuthMethodsResponse(await self._stream.read_exact(2))
        res.validate(request=req)
        return res.auth_method

    async def _socks_connect(self):
        req = ConnectRequest(
            host=self._dest_host,
            port=self._dest_port,
            rdns=self._rdns,
        )

        if req.need_resolve:
            _, addr = await self._resolver.resolve(req.host, family=socket.AF_UNSPEC)
            req.set_resolved_host(addr)

        await self._stream.write_all(bytes(req))

        res = ConnectResponse(await self._stream.read_exact(3))
        res.validate()

        # read remaining data (bind address)
        await self._stream.read()



import enum
import ipaddress
import logging
import socket

from .base_proxy import BaseProxyHandler, ProxyError
from .resolver import Resolver
from ..helpers import is_ipv4_address, is_ipv6_address

RSV = NULL = 0x00
SOCKS_VER4 = 0x04


class Command(enum.IntEnum):
    CONNECT = 0x01
    BIND = 0x02


class ReplyCode(enum.IntEnum):
    REQUEST_GRANTED = 0x5A
    REQUEST_REJECTED_OR_FAILED = 0x5B
    CONNECTION_FAILED = 0x5C
    AUTHENTICATION_FAILED = 0x5D


ReplyMessages = {
    ReplyCode.REQUEST_GRANTED: 'Request granted',
    ReplyCode.REQUEST_REJECTED_OR_FAILED: 'Request rejected or failed',
    ReplyCode.CONNECTION_FAILED: 'Request rejected because SOCKS server '
                                 'cannot connect to identd on the client',
    ReplyCode.AUTHENTICATION_FAILED: 'Request rejected because '
                                     'the client program '
                                     'and identd report different user-ids'
}


class Socks4ProxyHandler(BaseProxyHandler):
    def __init__(self, request, client_address, server, *, username=None):
        self.username = username
        self.logger = logging.getLogger(__name__)
        self.resolver = Resolver()

        super().__init__(request, client_address, server)

    def connect_to_remote(self) -> socket.socket:
        client = self.request

        version, command = client.recv(2)

        if version != SOCKS_VER4:
            self.respond(ReplyCode.REQUEST_REJECTED_OR_FAILED)
            raise ProxyError('Invalid socks version')

        if command != Command.CONNECT:
            self.respond(ReplyCode.REQUEST_REJECTED_OR_FAILED)
            raise ProxyError('Unsupported command')

        port = int.from_bytes(client.recv(2), 'big')
        host_bytes = client.recv(4)

        include_hostname = host_bytes[:3] == bytes([NULL, NULL, NULL])

        user = self.read_until_null().decode('ascii')
        if self.username and self.username != user:
            self.respond(ReplyCode.AUTHENTICATION_FAILED)
            raise ProxyError('Authentication failed')

        if include_hostname:
            host = self.read_until_null().decode('ascii')
        else:
            host = str(ipaddress.IPv4Address(host_bytes))

        self.logger.info('CONNECT {}:{}'.format(host, port))

        try:
            # remote = socket.create_connection((host, port))
            family, host = self.resolve_target_host(host)

            remote = socket.socket(
                family=family,
                type=socket.SOCK_STREAM
            )

            remote.connect((host, port))
        except Exception as e:
            self.logger.error(e)
            self.respond(ReplyCode.CONNECTION_FAILED)
            raise ProxyError('Connection refused by destination host') from e
        else:
            self.respond(ReplyCode.REQUEST_GRANTED)
            return remote

    def read_until_null(self) -> bytes:
        data = bytearray()
        while True:
            byte = ord(self.request.recv(1))
            if byte == NULL:
                break
            data.append(byte)
        return data

    def respond(self, code: ReplyCode):
        self.request.sendall(bytes([
            RSV,
            code,
            NULL,
            NULL,
            NULL,
            NULL,
            NULL,
            NULL,
        ]))

    def resolve_target_host(self, host):
        if is_ipv4_address(host):
            return socket.AF_INET, host
        if is_ipv6_address(host):
            return socket.AF_INET6, host
        return self.resolver.resolve(host=host)
